import { fallbackProperty, filterProperty, headersProperty, ImageCacheItBase, imageSourceProperty, loadModeProperty, overlayColorProperty, Priority, priorityProperty, srcProperty, stretchProperty, tintColorProperty } from './common';
import { Color, knownFolders, path as fsPath, ImageSource, Trace, Utils } from '@nativescript/core';
export class ImageCacheItError extends Error {
    static fromNative(native, message = undefined) {
        const error = new ImageCacheItError(message);
        error._native = native;
        return error;
    }
    get native() {
        return this._native;
    }
    get message() {
        if (!this._message) {
            this._message = this.native?.localizedDescription;
        }
        return this._message;
    }
    intoNative() {
        if (!this._native) {
            const exception = NSException.exceptionWithNameReasonUserInfo(NSGenericException, this.message, null);
            const info = {};
            info['ExceptionName'] = exception.name;
            info['ExceptionReason'] = exception.reason;
            info['ExceptionCallStackReturnAddresses'] = exception.callStackReturnAddresses;
            info['ExceptionCallStackSymbols'] = exception.callStackSymbols;
            info['ExceptionUserInfo'] = exception.userInfo;
            const error = NSError.alloc().initWithDomainCodeUserInfo('NativeScript', 1000, info);
            return error;
        }
        return this._native;
    }
}
export class ImageCacheIt extends ImageCacheItBase {
    constructor() {
        super();
        this._imageSourceAffectsLayout = true;
    }
    static get maxDiskCacheSize() {
        return NSCImageCacheItView.maxDiskCacheSize;
    }
    static set maxDiskCacheSize(size) {
        NSCImageCacheItView.maxDiskCacheSize = size;
    }
    static set maxMemoryCacheSize(size) {
        NSCImageCacheItView.maxMemoryCacheSize = size;
    }
    static get maxMemoryCacheSize() {
        return NSCImageCacheItView.maxMemoryCacheSize;
    }
    static get maxDiskCacheAge() {
        return NSCImageCacheItView.maxDiskCacheAge;
    }
    static set maxDiskCacheAge(age) {
        NSCImageCacheItView.maxDiskCacheAge = age;
    }
    createNativeView() {
        return NSCImageCacheItView.new();
    }
    initNativeView() {
        super.initNativeView();
        const nativeView = this.nativeViewProtected;
        nativeView.loadStart = (src) => {
            this.notify({ eventName: ImageCacheItBase.onLoadStartEvent, object: this, src });
        };
        nativeView.onLoadEnd = (src, image) => {
            this.notify({ eventName: ImageCacheItBase.onLoadEndEvent, object: this, src, image });
        };
        nativeView.onError = (src, error) => {
            this.notify({ eventName: ImageCacheItBase.onErrorEvent, object: this, src, error });
        };
        nativeView.onProgress = (current, total, progress, src) => {
            this.notify({ eventName: ImageCacheItBase.onProgressEvent, object: this, current, total, progress, src });
        };
        this.nativeViewProtected.runLayout = () => {
            this.requestLayout();
        };
    }
    onMeasure(widthMeasureSpec, heightMeasureSpec) {
        const width = Utils.layout.getMeasureSpecSize(widthMeasureSpec);
        const widthMode = Utils.layout.getMeasureSpecMode(widthMeasureSpec);
        const height = Utils.layout.getMeasureSpecSize(heightMeasureSpec);
        const heightMode = Utils.layout.getMeasureSpecMode(heightMeasureSpec);
        const nativeWidth = this.imageSource ? Utils.layout.toDevicePixels(this.imageSource.width) : 0;
        const nativeHeight = this.imageSource ? Utils.layout.toDevicePixels(this.imageSource.height) : 0;
        let measureWidth = Math.max(nativeWidth, this.effectiveMinWidth);
        let measureHeight = Math.max(nativeHeight, this.effectiveMinHeight);
        const finiteWidth = widthMode !== Utils.layout.UNSPECIFIED;
        const finiteHeight = heightMode !== Utils.layout.UNSPECIFIED;
        this._imageSourceAffectsLayout = widthMode !== Utils.layout.EXACTLY || heightMode !== Utils.layout.EXACTLY;
        if (nativeWidth !== 0 && nativeHeight !== 0 && (finiteWidth || finiteHeight)) {
            const scale = ImageCacheIt.computeScaleFactor(width, height, finiteWidth, finiteHeight, nativeWidth, nativeHeight, this.stretch);
            const resultW = Math.round(nativeWidth * scale.width);
            const resultH = Math.round(nativeHeight * scale.height);
            measureWidth = finiteWidth ? Math.min(resultW, width) : resultW;
            measureHeight = finiteHeight ? Math.min(resultH, height) : resultH;
            if (Trace.isEnabled()) {
                Trace.write('ImageCacheIt stretch: ' + this.stretch + ', nativeWidth: ' + nativeWidth + ', nativeHeight: ' + nativeHeight, Trace.categories.Layout);
            }
        }
        const widthAndState = ImageCacheIt.resolveSizeAndState(measureWidth, width, widthMode, 0);
        const heightAndState = ImageCacheIt.resolveSizeAndState(measureHeight, height, heightMode, 0);
        this.setMeasuredDimension(widthAndState, heightAndState);
    }
    static computeScaleFactor(measureWidth, measureHeight, widthIsFinite, heightIsFinite, nativeWidth, nativeHeight, imageStretch) {
        let scaleW = 1;
        let scaleH = 1;
        if ((imageStretch === 'aspectFill' || imageStretch === 'aspectFit' || imageStretch === 'fill') && (widthIsFinite || heightIsFinite)) {
            scaleW = nativeWidth > 0 ? measureWidth / nativeWidth : 0;
            scaleH = nativeHeight > 0 ? measureHeight / nativeHeight : 0;
            if (!widthIsFinite) {
                scaleW = scaleH;
            }
            else if (!heightIsFinite) {
                scaleH = scaleW;
            }
            else {
                // No infinite dimensions.
                switch (imageStretch) {
                    case 'aspectFit':
                        scaleH = scaleW < scaleH ? scaleW : scaleH;
                        scaleW = scaleH;
                        break;
                    case 'aspectFill':
                        scaleH = scaleW > scaleH ? scaleW : scaleH;
                        scaleW = scaleH;
                        break;
                }
            }
        }
        return { width: scaleW, height: scaleH };
    }
    [headersProperty.getDefault]() {
        return new Map();
    }
    [headersProperty.setNative](value) {
        if (this.nativeView && value) {
            this.nativeViewProtected.headers = value;
        }
    }
    getFileName(path) {
        let fileName = typeof path === 'string' ? path.trim() : '';
        if (fileName.indexOf('~/') === 0) {
            fileName = fsPath.join(knownFolders.currentApp().path, fileName.replace('~/', ''));
        }
        return fileName;
    }
    [fallbackProperty.setNative](src) {
        if (src === undefined || src === null) { }
        if (src instanceof ImageSource) {
            this.nativeViewProtected.fallback = src.ios;
        }
        else if (src instanceof UIImage) {
            this.nativeViewProtected.fallback = src;
        }
        else if (typeof src === 'string') {
            this.nativeViewProtected.fallback = ImageSource.fromFileOrResourceSync(this.getFileName(src))?.ios ?? null;
        }
    }
    [srcProperty.setNative](src) {
        if (!this.nativeViewProtected) {
            return;
        }
        if (typeof src === 'string') {
            if (src.startsWith('~/')) {
                this.nativeViewProtected.src = this.getFileName(src);
            }
            else {
                this.nativeViewProtected.src = src;
            }
        }
    }
    [loadModeProperty.setNative](value) {
        if (!this.nativeViewProtected) {
            return;
        }
        switch (value) {
            case 'sync':
                this.nativeViewProtected.loadMode = 0 /* NSCImageCacheItLoadMode.Sync */;
                break;
            case 'async':
            default:
                this.nativeViewProtected.loadMode = 1 /* NSCImageCacheItLoadMode.Async */;
                break;
        }
    }
    [stretchProperty.getDefault]() {
        return 'aspectFit';
    }
    [stretchProperty.setNative](value) {
        switch (value) {
            case 'aspectFit':
                this.nativeViewProtected.stretch = 1 /* NSCImageCacheItStretch.AspectFit */;
                break;
            case 'aspectFill':
                this.nativeViewProtected.stretch = 2 /* NSCImageCacheItStretch.AspectFill */;
                break;
            case 'fill':
                this.nativeViewProtected.stretch = 3 /* NSCImageCacheItStretch.Fill */;
                break;
            case 'none':
            default:
                this.nativeViewProtected.stretch = 0 /* NSCImageCacheItStretch.None */;
                break;
        }
    }
    [filterProperty.setNative](filter) {
        this.nativeViewProtected.filter = filter;
    }
    [tintColorProperty.setNative](value) {
        if (typeof value === 'string') {
            this.nativeViewProtected.imageTint = new Color(value).ios;
        }
        else {
            this.nativeViewProtected.imageTint = value?.ios ?? null;
        }
    }
    [overlayColorProperty.setNative](value) {
        if (typeof value === 'string') {
            this.nativeViewProtected.overlayColor = new Color(value).ios;
        }
        else {
            this.nativeViewProtected.imageTint = value?.ios ?? null;
        }
    }
    [priorityProperty.setNative](value) {
        if (!this.nativeViewProtected) {
            return;
        }
        switch (value) {
            case Priority.High:
                this.nativeViewProtected.priority = 2 /* NSCImageCacheItPriority.High */;
                break;
            case Priority.Low:
                this.nativeViewProtected.priority = 0 /* NSCImageCacheItPriority.Low */;
                break;
            default:
                this.nativeViewProtected.priority = 1 /* NSCImageCacheItPriority.Normal */;
                break;
        }
    }
    [imageSourceProperty.setNative](value) {
        if (this.nativeViewProtected && value?.ios) {
            this.nativeViewProtected.setImageSource(value.ios);
        }
    }
    static hasItem(src) {
        return new Promise((resolve, reject) => {
            NSCImageCacheItView.hasItem(src, (has) => {
                if (has) {
                    resolve(undefined);
                }
                else {
                    reject();
                }
            });
        });
    }
    static deleteItem(src) {
        return new Promise((resolve, reject) => {
            NSCImageCacheItView.deleteItem(src, (error) => {
                if (error) {
                    reject(ImageCacheItError.fromNative(error));
                }
                else {
                    resolve(undefined);
                }
            });
        });
    }
    static getItem(src, headers) {
        return new Promise((resolve, reject) => {
            NSCImageCacheItView.getItem(src, headers ?? {}, (error, data) => {
                if (error) {
                    reject(ImageCacheItError.fromNative(error));
                }
                else {
                    resolve(data);
                }
            });
        });
    }
    static clear() {
        return new Promise((resolve, reject) => {
            NSCImageCacheItView.clear(() => {
                resolve(undefined);
            });
        });
    }
    static enableAutoMM() {
        ImageCacheIt.enableAutoMM();
    }
    static disableAutoMM() {
        ImageCacheIt.disableAutoMM();
    }
}
//# sourceMappingURL=index.ios.js.map